#include <stdio.h>
#include <malloc.h>
#include "mdll.h"

void (*childfunc1) (void);
void (*childfunc2) (void);

void parentfunc1 (void) {
	puts ("Hello from parent1, called from child.");
}

void parentfunc2 (void) {
	puts ("Hello from parent2, called from child.");
}

static int dllApiIter (void * void_mdllH, char * name) {
void * p;

	p = MDLL_Import ((mdllHandle)void_mdllH, name);
	printf ("%s\t%p\n", name, p);
	return 0;
}

void main (void) {
	char * errStr;
        mdllHandle mdllH1, mdllH2;
	static struct mdllInit mdllCfg = {malloc, free};

	MDLL_Init (&mdllCfg);

	mdllH1 = MDLL_NewInterface ();
	mdllH2 = MDLL_NewInterface ();

	// export the 'parentfunc()' so the child can use it
	MDLL_Export (mdllH1, "parentfunc()", parentfunc1);
	MDLL_Export (mdllH1, "dllwalk()", MDLL_WalkPool);
	MDLL_Export (mdllH2, "parentfunc()", parentfunc2);
	MDLL_Export (mdllH2, "dllwalk()", MDLL_WalkPool);

	// dynamicly link the .dlls
	if (MDLL_Bind (mdllH1, "child.dll")) {
		errStr = MDLL_GetErr (mdllH1);
		if (errStr) {
			printf ("Error loading PE format: %s\n", errStr);
		}
		return;
	}

	if (MDLL_Bind (mdllH2, "child.dll")) {
		errStr = MDLL_GetErr (mdllH2);
		if (errStr) {
			printf ("Error loading PE format: %s\n", errStr);
		}
		return;
	}


	MDLL_WalkPool (mdllH1, dllApiIter, (void *)mdllH1);
	MDLL_WalkPool (mdllH2, dllApiIter, (void *)mdllH2);

	// import a child functions
	childfunc1 = MDLL_Import (mdllH1, "childfunc()");
	childfunc2 = MDLL_Import (mdllH2, "childfunc()");

	// execute that function
	if (childfunc1 != NULL) childfunc1 ();

	// Remove the "parentfunc()" exported API object
	MDLL_RemoveExport (mdllH1, "parentfunc()");

	// and free the memory used by the dll
	MDLL_Unbind (mdllH1);
	MDLL_DestroyInterface (mdllH1);

	if (childfunc2 != NULL) childfunc2 ();

	// Remove all exported object
	MDLL_RemoveExport (mdllH2, NULL);

	// and free the memory used by the dll
	MDLL_Unbind (mdllH2);
	MDLL_DestroyInterface (mdllH2);
}
